#ifndef INCLUDED_TRACK
#define INCLUDED_TRACK

#include "global.h"
#include "list.h"
extern "C" double atan2(double y, double x);


//------------------------------------------------------------------------------------------------
//	Track = a list of chunk
//------------------------------------------------------------------------------------------------

enum {LIN_TIME, LIN_VAL, LIN_ALL, SMOOTH_ALL,

	  VAL_LEFT = 0x0010, VAL_VALUE = 0x0020, VAL_RIGHT = 0x0040, VAL_MASK = 0x00F0,
	  OLD_LEFT = 0x0100, OLD_VALUE = 0x0200, OLD_RIGHT = 0x0400,

	  NEW_KEY = 0x10000, TEST_ONLY = 0x20000 };


class Chunk {
protected:
	// DATA
	int startTime, endTime;			// time
	int leftTime, rightTime;		// for acceleration/deceleration
	double startValue, endValue;	// initial and final value
	double leftValue, rightValue;	// value for bezier interpolation
	int isSynthetic;

	// FRIENDS
	friend class ViewTrack;
	friend class Track;
	friend class Routine;
	friend class Key;
	friend class Figure;

public:
	// CREATORS
	Chunk(int start, int end, double val0, double val1, 
		  double left, double right, int tLeft, int tRight, int synthetic = FALSE);
	~Chunk() {}

	// MANIPULATORS
	void SetLinear();
	void SetLinearValues();
	void SetLinearTimes();
	void SetLinearLeft();
	void SetLinearRight();
	void Smooth(class Chunk *prev, int modifyPrev, int modifyCurrent);

	// ACCESSORS
	double operator[](int time) const;
	void Draw() const;
	void DrawHandle() const;
	void Save(FILE *out) const;

#ifdef DEBUG
	void Dump(int full = FALSE) const {
		if (full) fprintf(stderr, " (%8g@%5d %8s(%8g,%8g, %5d,%5d) %8g@%5d)", startValue,startTime, isSynthetic ? "<bezier>":"bezier", leftValue,rightValue,leftTime,rightTime, endValue,endTime);
		else fprintf(stderr, " %g@%d %s(%g,%g, %d,%d)",startValue,startTime, isSynthetic ? "<bezier>":"bezier", leftValue,rightValue,leftTime,rightTime); }
	void DumpEnd() const { fprintf(stderr, " %g@%d", endValue, endTime); }
#endif
};


//------------------------------------------------------------------------------------------------
//	A track = a list of chunks
//------------------------------------------------------------------------------------------------

class Track {
	// DATA
	class List<Chunk> *data;
	class ListIterator<Chunk> current;

	// FRIENDS
	friend class ViewTrack;
	friend class Routine;
	friend class Figure;
#ifdef DEBUG
	friend class Interpolator;
	friend class Key;
#endif

public:
	// CREATORS
	Track() : data(new List<Chunk>), current(data) {}
	~Track() { if (data) data->free(); delete data; }

	// MANIPULATORS
	int AddInPlace(class Chunk *c);
	int AddInPlace(class Track *t);
	void AdjustNbBeats(int nbBeats);
	int ApplyModif(class Modif *m, class List<Modif> *modifs);
	void Delete(int start, int end, int nbBeats);
	double JointValue(int start=0, int end=1000000000);
	void RemoveSynthetic();
	void Smooth(int start, int end, int correct = FALSE);
	void TranslateTime(int offset, int start=0, int end=1000000000);
	double TranslateValue(double offset, int start=0, int end=1000000000);
	int Update(int time, double newValue, int nbBeats, int mode, int lastTime=0);

	// ACCESSORS
	int IsUsed() const { return (current); }
	int LastDefinedTime(int time) const;
	double operator[](int time);	// not const because 'current' modified

private:
	int FindPlace(int time);
};


//------------------------------------------------------------------------------------------------
//	An interpolator = a track for each joint
//------------------------------------------------------------------------------------------------

struct Interpolator {
	// DATA
	class Track track[NB_DOF];

	// CREATORS
	Interpolator() {}
	~Interpolator() {}

	// MANIPULATORS
	void SetBody(class Body *b, int time, int joint = -1);

	void AdjustNbBeats(int nbBeats);
	void RemoveSynthetic();
	void Smooth(int start, int end, int correct = FALSE);
	void TranslateTime(int offset, int start=0, int end=1000000000);

#ifdef DEBUG
	void Dump(int full=FALSE) const;
#endif
};


#endif
