#include "togl.h"
#include <GL/glu.h>
#include <math.h>	// for rint

#include "list.h"
#include "view.h"
#include "body.h"


class List<Light> *View3D::lights = new List<Light>;


//================================================================================================

#include "view3D.icc"

//------------------------------------------------------------------------------------------------

View3D::View3D(Togl *t)
{
	togl = t;
	forceWire = jointsOnly = FALSE;
	withRX = withRY = withRZ = TRUE;
	setupLight = TRUE;
	sphere = sphereX = sphereZ = disk = drawLocalAxis = FALSE;
	ortho = PERSP;
	drawFloor = drawAxis = TRUE;
	size = 1.0;
	bodies = new List<Body>;
	// register the view for the light
	Light::views->append(togl);
}

View3D::~View3D()
{
	// Remove cross references for the bodies in the view
	for(ListIterator<Body> b(bodies); b; ++b) {
		b()->views->remove(togl);
	}

	delete bodies; 

	// unregister the view for the light
	Light::views->remove(togl);
}

//------------------------------------------------------------------------------------------------

int View3D::Add(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
struct Tcl_CmdInfo info;
struct Body *body;
struct Light *light;
char *name;


	if (objc < 3) {
		Tcl_WrongNumArgs(interp, 2, objv, "body ?body ...?");
		return TCL_ERROR;
	}

	Tcl_ResetResult(interp);

	for(int a=2; a<objc; a++) {
		name = Tcl_GetStringFromObj(objv[a], NULL);
		if (Tcl_GetCommandInfo(interp, name, &info) == 0) {
			Tcl_AppendResult(interp, "no object named \"", name, "\"", (char *) NULL);
			return TCL_ERROR;
		}

		if (*name == 'b' && *(name+2) == 'd') { // body
			body = (class Body *) info.objClientData;
			bodies->append(body);
			body->views->append(togl);

			Tcl_AppendResult(interp, name, " ", (char *) NULL);
		}
	}

	Togl_PostRedisplay(togl);
	return TCL_OK;
}

int View3D::Remove(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
struct Tcl_CmdInfo info;
struct Body *body;
struct Light *light;
char *name;

	if (objc < 3) {
		Tcl_WrongNumArgs(interp, 2, objv, "body ?body ...?");
		return TCL_ERROR;
	}

	Tcl_ResetResult(interp);

	for(int a=2; a<objc; a++) {
		name = Tcl_GetStringFromObj(objv[a], NULL);
		if (Tcl_GetCommandInfo(interp, name, &info) == 0) {
			Tcl_AppendResult(interp, "no object named \"", name, "\"", (char *) NULL);
			return TCL_ERROR;
		}

		if (*name == 'b') { // body
			body = (class Body *) info.objClientData;
			bodies->remove(body);
			body->views->remove(togl);
		}

		Tcl_AppendResult(interp, name, " ", (char *) NULL);
	}

	Togl_PostRedisplay(togl);
	return TCL_OK;
}

//================================================================================================


void View3D::SetGLView(int picking, int x, int y) const
{
int width = Togl_Width(togl);
int height = Togl_Height(togl);
float aspect = (float) width / (float) height;
GLint viewport[4];


	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (picking) {
	    glGetIntegerv(GL_VIEWPORT, viewport);
	    gluPickMatrix((GLdouble) x, (GLdouble) (viewport[3] - y), 2.0, 2.0, viewport);
	}
	if (ortho != PERSP) {
		glScaled(1.0, aspect, -2.0/(viewpoint.zFar-viewpoint.zNear));
		glScaled(viewpoint.dist, viewpoint.dist, 1.0);
	} else {
		gluPerspective(viewpoint.fovy, aspect, viewpoint.zNear, viewpoint.zFar);
	}
	glTranslated(viewpoint.xTrans, viewpoint.yTrans, 0.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();						// Reset modelview matrix to the identity matrix



	// setup light
	if (setupLight) {
		class View3D *self = const_cast<View3D *>(this); 
		self->setupLight = FALSE;
		GLfloat p[4];
		p[0] = p[1] = p[3] = 0.0; p[2] = 1.0;
	    glLightfv(GL_LIGHT0, GL_POSITION, p);
		p[0] = p[1] = p[2] = p[3] = 0.0;
		glLightfv(GL_LIGHT0, GL_AMBIENT, p);
		p[0] = p[1] = p[2] = p[3] = 1.0;
		glLightfv(GL_LIGHT0, GL_DIFFUSE, p);
		glEnable(GL_LIGHT0);
	}
/*
	if (!picking && Light::changed) {
		int i = GL_LIGHT0;
		GLfloat p[4];

		for(ListIterator<Light> l(lights); l; ++l) {
			if (i <= GL_LIGHT7) {
				p[0] = l()->position[0]; p[1] = l()->position[1];
				p[2] = l()->position[2]; p[3] = l()->position[3]; 
			    glLightfv((GLenum) i, GL_POSITION, p);
				p[0] = l()->ambient[0]; p[1] = l()->ambient[1];
				p[2] = l()->ambient[2]; p[3] = l()->ambient[3]; 
			    glLightfv((GLenum) i, GL_AMBIENT, p);
				p[0] = l()->diffuse[0]; p[1] = l()->diffuse[1];
				p[2] = l()->diffuse[2]; p[3] = l()->diffuse[3]; 
			    glLightfv((GLenum) i, GL_DIFFUSE, p);
				glEnable((GLenum) i++);
			}
		}
		for (; i <= GL_LIGHT7; i++)
			glDisable((GLenum) i);

		Light::changed = FALSE;
	}
*/

	if (ortho == PERSP) {
		glTranslated(0.0, 0.0, -viewpoint.dist);
	}
	glRotated(viewpoint.elevation, 1.0, 0.0, 0.0);
	glRotated(viewpoint.azimut, 0.0, 1.0, 0.0);  		// Rotate by X, Y, and Z angles
	glRotated(viewpoint.twist, 0.0, 0.0, 1.0);

	if (!picking) {
		glDisable(GL_LIGHTING);
		if (drawAxis) {
			glPushMatrix();
			glScalef(3.0,3.0,3.0);
			glCallList(OBJ_AXIS);
			glPopMatrix();
		}
	}

	glTranslated(-viewpoint.xPos, -viewpoint.yPos, -viewpoint.zPos);
}


void View3D::Redisplay() const
{
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	SetGLView(FALSE, 0, 0);

	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	if (drawFloor)
		floor.Draw();		// draw the floor

	for(ListIterator<Body> b(bodies); b; ++b) {
		glPushMatrix();
		b()->Draw(forceWire ? FORCE_WIRE : WIRE, jointsOnly, withRX, withRY, withRZ);
		glPopMatrix();
	}

	if (sphere || sphereX || sphereZ || disk) {
		glPushMatrix();
		glDisable(GL_LIGHTING);
		glLoadMatrixd(matrix);
		glColor3f(1.0, 0.0, 0.0);
		glScalef(size, size, size);
		glCallList(sphere ? OBJ_SPHERE_XZ :
				   sphereX ? OBJ_SPHERE_X_ONLY :
				   sphereZ ? OBJ_SPHERE_Z_ONLY :
							 OBJ_DISK);
		glPopMatrix();
	} else if (drawLocalAxis) {
		glPushMatrix();
		glDisable(GL_LIGHTING);
		glTranslated(localAxis.X, localAxis.Y, localAxis.Z);

		if (fabs(viewpoint.elevation)>55) {
			glBegin(GL_LINES);
			  glColor3dv(localAxis.color1);
			  glVertex3f(-1.0, 0.0, 0.0);
			  glVertex3f(1.0, 0.0, 0.0);
			  glVertex3f(0.0, 0.0, -1.0);
			  glVertex3f(0.0, 0.0, 1.0);
			  glColor3dv(localAxis.color2);
			  glVertex3f(0.0, -1.0, 0.0);
			  glVertex3f(0.0, 1.0, 0.0);
			glEnd();
		} else {
//fprintf(stderr, "az = %f  --  %f %d\n", viewpoint.azimut, viewpoint.azimut/90.0, ((int) (viewpoint.azimut/90.0+0.5)));
			if (((int) (viewpoint.azimut/90.0+0.5))%2) {
				glBegin(GL_LINES);
				  glColor3dv(localAxis.color1);
				  glVertex3f(0.0, -1.0, 0.0);
				  glVertex3f(0.0, 1.0, 0.0);
				  glVertex3f(0.0, 0.0, -1.0);
				  glVertex3f(0.0, 0.0, 1.0);
				  glColor3dv(localAxis.color2);
				  glVertex3f(-1.0, 0.0, 0.0);
				  glVertex3f(1.0, 0.0, 0.0);
				glEnd();
			} else {
				glBegin(GL_LINES);
				  glColor3dv(localAxis.color1);
				  glVertex3f(-1.0, 0.0, 0.0);
				  glVertex3f(1.0, 0.0, 0.0);
				  glVertex3f(0.0, -1.0, 0.0);
				  glVertex3f(0.0, 1.0, 0.0);
				  glColor3dv(localAxis.color2);
				  glVertex3f(0.0, 0.0, -1.0);
				  glVertex3f(0.0, 0.0, 1.0);
				glEnd();
			}
		}
		glPopMatrix();	
	}
}


int View3D::Pick(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int joint) const
{
GLuint selectBuf[SELECT_BUFSIZE], *ptr, names, min = 0xffffffff;
GLint hits;
GLint viewport[4];
int x, y, i, body = -1, part = -1;
Tcl_Obj *res;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "x y");
		return TCL_ERROR;
	}

	// get the coordinate of the pick
	Tcl_GetIntFromObj(interp, objv[2], &x);
	Tcl_GetIntFromObj(interp, objv[3], &y);

	// enter in selection mode
    glSelectBuffer(SELECT_BUFSIZE, selectBuf);
    glRenderMode(GL_SELECT);

    glInitNames();
    glPushName(0);

	SetGLView(TRUE, x, y);

	glEnable(GL_DEPTH_TEST);
	for(ListIterator<Body> b(bodies); b; ++b) {
		if (b()->ghost == BODY) {
			glPushMatrix();
			glLoadName(b()->number);	// name of the body
			glPushName(0);
			b()->Draw(SELECT, jointsOnly, withRX, withRY, withRZ);
			glPopName();
			glPopMatrix();
		}
	}
    glFlush();
    hits = glRenderMode(GL_RENDER);

	// analyze the result
    ptr = selectBuf;
    for (i = 0; i < hits; i++) {
		if ((names = *ptr++) != 2) {
			Tcl_SetResult(interp, "[view3D::pick] internal error", NULL);
			return TCL_ERROR;
		}
		ptr++;	// skip the max
		if (*ptr < min) {
			min = *ptr++;
			body = *ptr++;
			part = *ptr++;
		} else {
			ptr += 3; // not good, skip it
		}
	}

	// convert to name
	res = Tcl_NewListObj(0, NULL);
	if (part >= OBJ_PART && part < OBJ_PART+OBJ_NB_PART) {
		extern int selectedPart;
		extern double *selectedPartMatrix;
		extern char *jointName[], *partName[];
		char tmp[10];

		// get coord of the base of the object
	    glMatrixMode(GL_MODELVIEW);
		selectedPart = part;
		selectedPartMatrix = (double *) matrix;
		for(ListIterator<Body> b(bodies); b; ++b) {
			if (b()->number == body) {
				b()->Draw(GETPOS, jointsOnly, withRX, withRY, withRZ);
				break;
			}
		}

		// get the matrix
		double proj[16], x, y, z;
		glGetDoublev(GL_PROJECTION_MATRIX, proj);
		gluProject(0,0,0, matrix, proj, viewport, &x, &y, &z);

		// set the result
		sprintf(tmp, "body%d", body);
		Tcl_ListObjAppendElement(interp, res, Tcl_NewStringObj(tmp, -1));
		Tcl_ListObjAppendElement(interp, res, Tcl_NewStringObj(
				joint ? jointName[part-OBJ_PART] : partName[part-OBJ_PART], -1));
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int View3D::Snapshot(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
char *snapData, header[50];
Tcl_Obj *res;
int j, w, h, oldW = 0, oldH = 0;

	// check if we specify the size
	if (objc > 2) {
		if (objc != 4) {
			Tcl_WrongNumArgs(interp, 2, objv, "?width height?");
			return TCL_ERROR;
		}
		Tcl_GetIntFromObj(interp, objv[2], &w);
		Tcl_GetIntFromObj(interp, objv[3], &h);
		oldW = togl->Width;
		oldH = togl->Height;
		togl->Width = w;
		togl->Height = h;
		Redisplay();
	}

	// get the snapshot data
	snapData = new char[3*4*4*(togl->Width+1)/4*(togl->Height+1)/4+1];
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadPixels(0,0, togl->Width, togl->Height,
				 GL_RGB, GL_UNSIGNED_BYTE, snapData);

	sprintf(header, "P6\n%d %d\n255\n", togl->Width, togl->Height);
	res = Tcl_NewStringObj(header, -1);
	for (j=togl->Height-1; j>=0; j--) {
		Tcl_AppendToObj(res, &snapData[j*togl->Width*3],
						togl->Width*3);
	}
	delete[] snapData;
	Tcl_SetObjResult(interp, res);

	// restore the state of the widget
	if (objc > 2) {
		togl->Width = oldW;
		togl->Height = oldH;
		Redisplay();
	}
	glPixelStorei(GL_PACK_ALIGNMENT, 4);

	return TCL_OK;
}


//================================================================================================

int View3D_ObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
struct Togl *togl = (struct Togl *)clientData;
struct View3D *view = (struct View3D *) Togl_GetClientData(togl);
static char *optionsPtr[] = {
	"configure", "render", "swapbuffers", "makecurrent", 
	"get", "set",
	"add", "remove", "pick", "pickjoint", "snapshot",
	"ortho", "persp", (char *) NULL};
enum {
	config, render, swap, current,
	get, set,
	add, remove, pick, pickjoint, snap,
	ortho, persp } index;


	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "?options?");
		return TCL_ERROR;
	}

	if (Tcl_GetIndexFromObj(interp, objv[1], optionsPtr, "option", 0, (int *) &index) != TCL_OK) {
    	return TCL_ERROR;
	}

	Togl_MakeCurrent(togl);

	switch (index) {
		case config:
		case render:
		case swap:
		case current:	return Togl_DefaultProcessing(togl, interp, index, objc, objv);

		case get:		return view->Get(interp, objc, objv);
		case set:		return view->Set(interp, objc, objv);
		case add:		return view->Add(interp, objc, objv);
		case remove:	return view->Remove(interp, objc, objv);
		case pick:		return view->Pick(interp, objc, objv, FALSE);
		case pickjoint:	return view->Pick(interp, objc, objv, TRUE);

		case snap:		return view->Snapshot(interp, objc, objv);

		case ortho:		view->ortho = ORTHO;
						view->Redisplay();
						Togl_PostRedisplay(togl);
						//Tk_SetClass(togl->TkWin, "View3DProjection");
						break;
		case persp:		view->ortho = PERSP;
						view->Redisplay();
						Togl_PostRedisplay(togl);
						//Tk_SetClass(togl->TkWin, "View3D");
						break;
	}

	return TCL_OK;
}
