#include "togl.h"
#include <GL/glu.h>
#include <math.h>

#include "list.h"
#include "movie.h"
#include "view.h"
#include "routine.h"	// for the .icc (variable movie)


GLuint ViewMovie::Font = 0;


//================================================================================================
//
//	Bezier3D
//
//================================================================================================

double Bezier3D::X(int time) const
{
double t = (double) (time-start)/((double) end-start), t2, t3, u, u2, u3;

	t2 = t*t;
	t3 = t2*t;
	u = 1.0 - t;
	u2 = u*u;
	u3 = u2*u;
	return startX*u3 + 3.0*(leftX*t*u2 + rightX*t2*u) + endX*t3;
}

double Bezier3D::Y(int time) const
{
double t = (double) (time-start)/((double) end-start), t2, t3, u, u2, u3;

	t2 = t*t;
	t3 = t2*t;
	u = 1.0 - t;
	u2 = u*u;
	u3 = u2*u;
	return startY*u3 + 3.0*(leftY*t*u2 + rightY*t2*u) + endY*t3;
}

double Bezier3D::Z(int time) const
{
double t = (double) (time-start)/((double) end-start), t2, t3, u, u2, u3;

	t2 = t*t;
	t3 = t2*t;
	u = 1.0 - t;
	u2 = u*u;
	u3 = u2*u;
	return startZ*u3 + 3.0*(leftZ*t*u2 + rightZ*t2*u) + endZ*t3;
}

//================================================================================================
//
//	Movie
//
//================================================================================================

Movie::Movie(char *name, List<Bezier3D> *eye, List<Bezier3D> *center) : 
	d_name(name), d_eye(eye), d_center(center)
{
}

Movie::~Movie()
{
	if (d_eye) {
		d_eye->free();
		delete d_eye;
	}

	if (d_center) {
		d_center->free();
		delete d_center;
	}
}

//------------------------------------------------------------------------------------------------

void Movie::NewName(char *name)
{
	delete[] d_name;
	d_name = new char[strlen(name)+1];
	strcpy(d_name, name);
}

//------------------------------------------------------------------------------------------------

#ifdef DEBUG
void Movie::Dump() const
{
ListIterator<Bezier3D> p;

	fprintf(stderr, "    name = '%s'\n", d_name);

	fprintf(stderr, "        eye = ");
	for(p.reset(d_eye); p; ++p)
		p()->Dump(TRUE);
	fprintf(stderr, "\n");

	fprintf(stderr, "        center = ");
	for(p.reset(d_center); p; ++p)
		p()->Dump(TRUE);
	fprintf(stderr, "\n");
}
#endif

//------------------------------------------------------------------------------------------------

void Movie::Save(FILE *out) const
{
ListIterator<Bezier3D> p;
extern void Stringify(FILE *out, char *str);


	fprintf(out, "movie(");
	Stringify(out, d_name);
	fprintf(out, ") = {\n");

	fprintf(out, "    center = ");
	for(p.reset(d_center); p; ++p) {
		fprintf(out, "<%g,%g,%g>@%d bezier(<%g,%g,%g>, <%g,%g,%g>) ",
				p()->startX,p()->startY,p()->startZ, p()->start,
				p()->leftX,p()->leftY,p()->leftZ, p()->rightX,p()->rightY,p()->rightZ);
		if (p.isTail()) {
			fprintf(out, "<%g,%g,%g>@%d",
					p()->endX,p()->endY,p()->endZ, p()->end);
		}
	}
	fprintf(out, "\n");

	fprintf(out, "    eye = ");
	for(p.reset(d_eye); p; ++p) {
		fprintf(out, "<%g,%g,%g>@%d bezier(<%g,%g,%g>, <%g,%g,%g>) ",
				p()->startX,p()->startY,p()->startZ, p()->start,
				p()->leftX,p()->leftY,p()->leftZ, p()->rightX,p()->rightY,p()->rightZ);
		if (p.isTail()) {
			fprintf(out, "<%g,%g,%g>@%d",
					p()->endX,p()->endY,p()->endZ, p()->end);
		}
	}
	fprintf(out, "\n}\n");
}


//================================================================================================

#include "viewMovie.icc"

//------------------------------------------------------------------------------------------------

ViewMovie::ViewMovie(Togl *t) : View3D(t)
{
	if (!Font) {
		Font = Togl_LoadBitmapFont(togl, TOGL_BITMAP_8_BY_13);
	}

	mode = EDIT;
	currentTime = askedTime = 0;
	showHeight = showHandle = showLine = FALSE;
	showPaths = TRUE;
	setupLight = TRUE;
	colorEye[0] = colorEye[1] = 1.0; colorEye[2] = colorEye[3] = 0.0; 
	colorCenter[0] = colorCenter[2] = 1.0; colorCenter[1] = colorCenter[3] = 0.0; 

	eye = new List<Bezier3D>;
	center = new List<Bezier3D>;
	//twist = new List<Bezier>;
}

ViewMovie::~ViewMovie()
{
}

//------------------------------------------------------------------------------------------------

int ViewMovie::Record(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int nb;
struct Tcl_Obj **list;
double azimut, elevation, twist, dist, centerX, centerY, centerZ, eyeX, eyeY, eyeZ;


	if (objc != 3) {
		Tcl_WrongNumArgs(interp, 2, objv, "record movieViewpoint");
		return TCL_ERROR;
	}

	// 1. get the viewpoint
	Tcl_ListObjGetElements(interp, objv[2], &nb, &list);
	if (nb != 7) {
		Tcl_AppendResult(interp, "bad viewpoint", (char *) NULL);
		return TCL_ERROR;
	}

	// 2. get the parameters & compute values
	Tcl_GetDoubleFromObj(interp, list[0], &azimut);
	azimut = azimut*3.141592654/180.0;
	Tcl_GetDoubleFromObj(interp, list[1], &elevation);
	elevation = elevation*3.141592654/180.0;
	Tcl_GetDoubleFromObj(interp, list[2], &twist);
	twist = twist*3.141592654/180.0;
	Tcl_GetDoubleFromObj(interp, list[3], &dist);
	Tcl_GetDoubleFromObj(interp, list[4], &centerX);
	Tcl_GetDoubleFromObj(interp, list[5], &centerY);
	Tcl_GetDoubleFromObj(interp, list[6], &centerZ);

	eyeX = centerX - sin(azimut)*cos(elevation)*dist;
	eyeY = centerY +             sin(elevation)*dist;
	eyeZ = centerZ + cos(azimut)*cos(elevation)*dist;

//fprintf(stderr, "@%d   <%g,%g,%g>  ----> <%g,%g,%g>\n", currentTime, eyeX,eyeY,eyeZ, centerX,centerY,centerZ);

	// 3. store data in the path
	ListIterator<Bezier3D> p;
	for(p.reset(eye); p; ++p) {
		if (p()->start == currentTime || p()->end == currentTime)
			break;
	}
	if (!p || askedTime > currentTime) {
		// new part
		Insert(eye, eyeX, eyeY, eyeZ);
		Insert(center, centerX, centerY, centerZ);
	} else {
		// move old part
		ApplyModif(eye, currentTime, HANDLE_TO_POINT, eyeX, eyeY, eyeZ);
		ApplyModif(eye, currentTime, SET_POINT, eyeX, eyeY, eyeZ);

		ApplyModif(center, currentTime, HANDLE_TO_POINT, centerX, centerY, centerZ);
		ApplyModif(center, currentTime, SET_POINT, centerX, centerY, centerZ);
	}

	Togl_PostRedisplay(togl);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

void ViewMovie::Insert(List<Bezier3D> *path, double X, double Y, double Z)
{
ListIterator<Bezier3D> p;


	for(p.reset(path); p; ++p) {
		if (p()->start < currentTime && currentTime < p()->end)
			break;
	}
	// new path
//fprintf(stderr, "insertion ::   c = %d     a = %d\n", askedTime, currentTime);
	if (askedTime == currentTime) {
		Bezier3D *b = new Bezier3D(currentTime, p()->end);
		// set values
		p()->end = currentTime;
		b->rightX = p()->rightX;
		b->rightY = p()->rightY;
		b->rightZ = p()->rightZ;
		b->endX = p()->endX;
		b->endY = p()->endY;
		b->endZ = p()->endZ;
		// insert it
		path->insert(p, b);
		// adjust it
		ApplyModif(path, currentTime, SET_POINT, X, Y, Z);
		ApplyModif(path, currentTime, SCALE_HANDLE, 0, 0, 0);

	} else {
		p.last(path);
		Bezier3D *b = new Bezier3D(p()->end, askedTime);
		b->startX = p()->endX;
		b->startY = p()->endY;
		b->startZ = p()->endZ;
		// insert it
		path->append(b);
		// adjust it
		ApplyModif(path, askedTime, SET_POINT, X, Y, Z);

		ApplyModif(path, p()->end, SCALE_HANDLE, 0, 0, 0);
		ApplyModif(path, askedTime, SCALE_HANDLE, 0, 0, 0);
	}

//fprintf(stderr, "path = "); for(p.reset(path); p; ++p) p()->Dump(TRUE); fprintf(stderr, "\n");
}

//------------------------------------------------------------------------------------------------

int ViewMovie::Modify(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int nb, time = 0, path, type;
struct Tcl_Obj **list;
double x,y,z;
static char *typePtr[] = {"point", "lefthandle", "righthandle", 
						  "align", "delete", "linearize", (char *) 0};
static char *pathPtr[] = {"eye", "center", (char *) 0};


	if (objc != 6) {
		Tcl_WrongNumArgs(interp, 2, objv, "{path time type} x y z");
		return TCL_ERROR;
	}

	// 1. get the element to modify
	Tcl_ListObjGetElements(interp, objv[2], &nb, &list);
	if (nb != 3) {
		Tcl_AppendResult(interp, "bad designation of element should be {path time type}",
						 (char *) NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetIndexFromObj(interp, list[0], pathPtr, "path", 0, &path) != TCL_OK) {
    	return TCL_ERROR;
	}
	Tcl_GetIntFromObj(interp, list[1], &time);
	if (Tcl_GetIndexFromObj(interp, list[2], typePtr, "type", 0, &type) != TCL_OK) {
    	return TCL_ERROR;
	}

	// 2. get the coords
	Tcl_GetDoubleFromObj(interp, objv[3], &x);
	Tcl_GetDoubleFromObj(interp, objv[4], &y);
	Tcl_GetDoubleFromObj(interp, objv[5], &z);

	// 3. do it
//fprintf(stderr, "modify@%d   type %d    <%g,%g,%g>\n",time,type, x,y,z);
	if (type == 3) {
		// Align
		AlignHandles(path == 0 ? eye : center, time);
	} else if (type == 4) {
		// Delete
		DeletePoint(eye, time);
		DeletePoint(center, time);
		currentEye.reset(eye);
		currentCenter.reset(center);
	} else if (type == 5) {
		// Linearize
		ApplyModif(path == 0 ? eye : center, time, SCALE_HANDLE, 0, 0, 0);
	} else {
		// Move an element
		if (type+1 == MOVE_POINT) {
			ApplyModif(path == 0 ? eye : center, time, LEFT_HANDLE, x, y, z);
			ApplyModif(path == 0 ? eye : center, time, RIGHT_HANDLE, x, y, z);
		}
		ApplyModif(path == 0 ? eye : center, time, type+1, x, y, z);
												// hack : see 'type+1' in view.h
	}

	Togl_PostRedisplay(togl);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

void ViewMovie::ApplyModif(List<Bezier3D> *path, int time, int type, double X, double Y, double Z)
{
	for(ListIterator<Bezier3D> p(path); p; ++p) {
		if (type == SET_POINT && time == p()->start) {
		// SET_POINT
			p()->startX = X;
			p()->startY = Y;
			p()->startZ = Z;
		} else if (type == SET_POINT && time == p()->end) {
			p()->endX = X;
			p()->endY = Y;
			p()->endZ = Z;
		} else if (type == MOVE_POINT && time == p()->start) {
		// MOVE_POINT
			p()->startX += X;
			p()->startY += Y;
			p()->startZ += Z;
		} else if (type == MOVE_POINT && time == p()->end) {
			p()->endX += X;
			p()->endY += Y;
			p()->endZ += Z;
		} else if (type == LEFT_HANDLE && time == p()->start) {
		// move LEFT_HANDLE
			p()->leftX += X;
			p()->leftY += Y;
			p()->leftZ += Z;
		} else if (type == RIGHT_HANDLE && time == p()->end) {
		// move RIGHT_HANDLE
			p()->rightX += X;
			p()->rightY += Y;
			p()->rightZ += Z;
		} else if (type == SCALE_HANDLE && time == p()->start) {
		// SCALE_HANDLE
			p()->leftX = p()->startX + (p()->endX - p()->startX)/3.0;
			p()->leftY = p()->startY + (p()->endY - p()->startY)/3.0;
			p()->leftZ = p()->startZ + (p()->endZ - p()->startZ)/3.0;
		} else if (type == SCALE_HANDLE && time == p()->end) {
			p()->rightX = p()->endX - (p()->endX - p()->startX)/3.0;
			p()->rightY = p()->endY - (p()->endY - p()->startY)/3.0;
			p()->rightZ = p()->endZ - (p()->endZ - p()->startZ)/3.0;
		} else if (type == HANDLE_TO_POINT && time == p()->start) {
		// HANDLE_TO_POINT
			p()->leftX += X - p()->startX;
			p()->leftY += Y - p()->startY;
			p()->leftZ += Z - p()->startZ;
		} else if (type == HANDLE_TO_POINT && time == p()->end) {
			p()->rightX += X - p()->endX;
			p()->rightY += Y - p()->endY;
			p()->rightZ += Z - p()->endZ;
		}
	}
}

//------------------------------------------------------------------------------------------------

void ViewMovie::AlignHandles(List<Bezier3D> *path, int time)
{
double pX = 0, pY = 0, pZ = 0;
Bezier3D *l = 0, *r = 0;

	// find the point & handles
	for(ListIterator<Bezier3D> p(path); p; ++p) {
		if (time == p()->start) {
			if (p.isHead()) {
				return;
			}
			r = p.prev();
			pX = p()->startX;
			pY = p()->startY;
			pZ = p()->startZ;
			l = p();
			break;
		} else if (time == p()->end) {
			if (p.isTail()) {
				return;
			}
			r = p();
			pX = p()->endX;
			pY = p()->endY;
			pZ = p()->endZ;
			l = p.next();
			break;
		}
	}

	if (!l) {
		return;
	}

	// compute the alignment
	double ldx, ldy, ldz, rdx, rdy, rdz, ltheta, rtheta, theta, lphi, rphi, phi, ldist, rdist;

	ldx = l->leftX - pX;
	ldy = l->leftY - pY;
	ldz = l->leftZ - pZ;

	rdx = pX - r->rightX;
	rdy = pY - r->rightY;
	rdz = pZ - r->rightZ;
//fprintf(stderr, "left = <%g,%g,%g>     right = <%g,%g,%g>\n", ldx,ldy,ldz, rdx,rdy,rdz);


	ldist = sqrt(ldx*ldx+ldy*ldy+ldz*ldz);
	lphi = atan2(ldy, sqrt(ldx*ldx+ldz*ldz)+0.00000001);
	ltheta = fabs(ldz)>0.00000001 ? atan2(ldx, ldz) : ldx < 0 ? -3.141592654/2 : 3.141592654/2;

	rdist = sqrt(rdx*rdx+rdy*rdy+rdz*rdz);
	rphi = atan2(rdy, sqrt(rdx*rdx+rdz*rdz)+0.00000001);
	rtheta = fabs(rdz)>0.00000001 ? atan2(rdx, rdz) : rdx < 0 ? -3.141592654/2 : 3.141592654/2;

	theta = (ltheta/ldist + rtheta/rdist)/(1/ldist+1/rdist);
	phi = (lphi/ldist + rphi/rdist)/(1/ldist+1/rdist);

//fprintf(stderr, "THETA   %g,%g    %%   %g,%g   ----->  %g\n", ldist,180.0/3.141592654*ltheta,rdist,180.0/3.141592654*rtheta,180.0/3.141592654*theta);
//fprintf(stderr, "PHI     %g,%g    %%   %g,%g   ----->  %g\n", ldist,180.0/3.141592654*lphi,rdist,180.0/3.141592654*rphi,180.0/3.141592654*phi);

	// store the new handles
	l->leftX = pX + cos(theta)*cos(phi)*ldist;
	l->leftY = pY +            sin(phi)*ldist;
	l->leftZ = pZ + sin(theta)*cos(phi)*ldist;

	r->rightX = pX - cos(theta)*cos(phi)*rdist;
	r->rightY = pY -            sin(phi)*rdist;
	r->rightZ = pZ - sin(theta)*cos(phi)*rdist;
}

//------------------------------------------------------------------------------------------------

void ViewMovie::DeletePoint(List<Bezier3D> *path, int time)
{
//fprintf(stderr, "deleting at %d on %x\n", time, path);

	for(ListIterator<Bezier3D> p(path); p; ++p) {
		if (time == p()->start || (p.isTail() && time == p()->end)) {
			if (p.isHead()) {
				return;
			} else {
//fprintf(stderr, "prev     "); p.prev()->Dump(TRUE);
//fprintf(stderr, "\ncurrent  "); p()->Dump(TRUE);
//fprintf(stderr, "\n");

				p.prev()->end = p()->end;

				p.prev()->rightX = p()->rightX;
				p.prev()->rightY = p()->rightY;
				p.prev()->rightZ = p()->rightZ;
				p.prev()->endX = p()->endX;
				p.prev()->endY = p()->endY;
				p.prev()->endZ = p()->endZ;

				path->remove(p());
				return;
			}
		}
	}
}

//================================================================================================

int ViewMovie::Viewpoint(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
double dx, dy, dz;

	if (!currentEye || mode != VIEW) {
		return TCL_OK;
	}

	viewpoint.xTrans = viewpoint.yTrans = 0.0;

	SetTime(currentTime);
	viewpoint.xPos = currentCenter()->X(currentTime);
	viewpoint.yPos = currentCenter()->Y(currentTime);
	viewpoint.zPos = currentCenter()->Z(currentTime);

	dx = currentEye()->X(currentTime) - viewpoint.xPos;
	dy = currentEye()->Y(currentTime) - viewpoint.yPos;
	dz = currentEye()->Z(currentTime) - viewpoint.zPos;
	viewpoint.dist = sqrt(dx*dx+dy*dy+dz*dz);

	viewpoint.elevation = 180.0/3.141592654*atan2(dy, sqrt(dx*dx+dz*dz)+0.00000001);
	viewpoint.azimut = fabs(dz)>0.00000001 ? -180.0/3.141592654*atan2(dx, dz) : 
											  dx < 0 ? 90.0 : -90.0;

	Togl_PostRedisplay(togl);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

void ViewMovie::SetGLView(int picking, int x, int y) const
{
ViewMovie *self = const_cast<ViewMovie *>(this); 
int width = Togl_Width(togl);
int height = Togl_Height(togl);
float aspect = (float) width / (float) height;
GLint viewport[4];
double dx = 0.0, dy = 0.0, dz = 0.0, data;


	// 1. init the view with the geometry
	glViewport(0, 0, width, height);

	// 2. set persp or ortho mode
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (picking) {
	    glGetIntegerv(GL_VIEWPORT, viewport);
	    gluPickMatrix((GLdouble) x, (GLdouble) (viewport[3] - y), 2.0, 2.0, viewport);
	}
	if (ortho != PERSP && mode == EDIT) {
		glScaled(1.0, aspect, -2.0/(viewpoint.zFar-viewpoint.zNear));
		glScaled(viewpoint.dist, viewpoint.dist, 1.0);
	} else {
		gluPerspective(viewpoint.fovy, aspect, viewpoint.zNear, viewpoint.zFar);
	}

	// 3. translate view if 
	if (mode != PLAY) {
		glTranslated(viewpoint.xTrans, viewpoint.yTrans, 0.0);
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();						// Reset modelview matrix to the identity matrix

	// 4. setup the lights (if needed)
	if (setupLight) {
		self->setupLight = FALSE;
		GLfloat p[4];
		p[0] = p[1] = p[3] = 0.0; p[2] = 1.0;
	    glLightfv(GL_LIGHT0, GL_POSITION, p);
		p[0] = p[1] = p[2] = p[3] = 0.0;
		glLightfv(GL_LIGHT0, GL_AMBIENT, p);
		p[0] = p[1] = p[2] = p[3] = 1.0;
		glLightfv(GL_LIGHT0, GL_DIFFUSE, p);
		glEnable(GL_LIGHT0);
	}
/*
	if (!picking && Light::changed) {
		int i = GL_LIGHT0;;
		GLfloat p[4];

		for(ListIterator<Light> l(lights); l; ++l) {
			if (i <= GL_LIGHT7) {
				p[0] = l()->position[0]; p[1] = l()->position[1];
				p[2] = l()->position[2]; p[3] = l()->position[3]; 
			    glLightfv((GLenum) i, GL_POSITION, p);
				p[0] = l()->ambient[0]; p[1] = l()->ambient[1];
				p[2] = l()->ambient[2]; p[3] = l()->ambient[3]; 
			    glLightfv((GLenum) i, GL_AMBIENT, p);
				p[0] = l()->diffuse[0]; p[1] = l()->diffuse[1];
				p[2] = l()->diffuse[2]; p[3] = l()->diffuse[3]; 
			    glLightfv((GLenum) i, GL_DIFFUSE, p);
				glEnable((GLenum) i++);
			}
		}
		for (; i <= GL_LIGHT7; i++)
			glDisable((GLenum) i);

		Light::changed = FALSE;
	}
*/

	// 5. setup the eye position
	if (mode != PLAY) {
		if (ortho == PERSP || mode == VIEW) {
			glTranslated(0.0, 0.0, -viewpoint.dist);
		}
		glRotated(viewpoint.elevation, 1.0, 0.0, 0.0);
		glRotated(viewpoint.azimut, 0.0, 1.0, 0.0);  		// Rotate by X, Y, and Z angles
		glRotated(viewpoint.twist, 0.0, 0.0, 1.0);
	} else {
		self->SetTime(currentTime);
		dx = currentEye()->X(currentTime) - currentCenter()->X(currentTime);
		dy = currentEye()->Y(currentTime) - currentCenter()->Y(currentTime);
		dz = currentEye()->Z(currentTime) - currentCenter()->Z(currentTime);
		data = sqrt(dx*dx+dy*dy+dz*dz);
//fprintf(stderr, "\npick = %s   ---  %g,%g,%g\n", picking?"yes":"no",dx,dy,dz);
//fprintf(stderr, "dist = %g\n", data);
		glTranslated(0.0, 0.0, -data);

		data = 180.0/3.141592654*atan2(dy, sqrt(dx*dx+dz*dz)+0.00000001);
//fprintf(stderr, "elevation = %g\n", data);
		glRotated(data, 1.0, 0.0, 0.0);	// elevation
		data = fabs(dz)>0.00000001 ? 180.0/3.141592654*atan2(dx, dz) : dx < 0 ? -90.0 : 90.0;
//fprintf(stderr, "azimut = %g\n", data);
		glRotated(-data, 0.0, 1.0, 0.0); // azimut
		glRotated(viewpoint.twist, 0.0, 0.0, 1.0);
	}

	if (!picking) {
		glDisable(GL_LIGHTING);
		if (drawAxis && mode == EDIT) {
			glPushMatrix();
			glScalef(3.0,3.0,3.0);
			glCallList(OBJ_AXIS);
			glPopMatrix();
		}
	}

	// 6. setup the center (point at which we are looking)
	if (mode != PLAY) {
		glTranslated(-viewpoint.xPos, -viewpoint.yPos, -viewpoint.zPos);
	} else {
		glTranslated(-currentCenter()->X(currentTime), -currentCenter()->Y(currentTime), 
					 -currentCenter()->Z(currentTime));
	}
}

//------------------------------------------------------------------------------------------------

void ViewMovie::Redisplay() const
{
ViewMovie *self = const_cast<ViewMovie *>(this); 
int start, end, delta;
char text[10];


	if (!currentEye) {
		glClearColor(0.0, 0.0, 0.0, 0.0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		return;
	}

	View3D::Redisplay();

	if (mode == PLAY || !showPaths) {
		return;
	}

	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	//---------------------------------------
	// 1. draw the path of the EYE and CENTER
	ListIterator<Bezier3D> pe, pc;
	for(pe.reset(eye), pc.reset(center); pe; ++pe, ++pc) {
		int t;

		// 1.1. get parameters
		start = pe()->start;
		end = pe()->end;
		delta = (end-start)/50;

		// 1.2. draw the bezier curve for EYE
		glColor3f(colorEye[0], colorEye[1], colorEye[2]);
		glBegin(GL_LINE_STRIP);
		glVertex3f(pe()->startX, pe()->startY, pe()->startZ);
		for(t=start; t<=end; t+=delta) {
			glVertex3f(pe()->X(t), pe()->Y(t), pe()->Z(t));
		}
		glEnd();

		// 1.3. draw the bezier curve for CENTER
		glColor3f(colorCenter[0], colorCenter[1], colorCenter[2]);
		glBegin(GL_LINE_STRIP);
		glVertex3f(pc()->startX, pc()->startY, pc()->startZ);
		for(t=start; t<=end; t+=delta) {
			glVertex3f(pc()->X(t), pc()->Y(t), pc()->Z(t));
		}
		glEnd();

		// 1.4. draw the points + handles
		glColor3f(colorEye[0], colorEye[1], colorEye[2]);
		glPointSize(4.0);
		glBegin(GL_POINTS);
		  glVertex3d(pe()->startX, pe()->startY, pe()->startZ);
		  if (pe.isTail()) {
			  glVertex3d(pe()->endX, pe()->endY, pe()->endZ);
		  }
		  if (showHandle) {
			glVertex3d(pe()->leftX, pe()->leftY, pe()->leftZ);
			glVertex3d(pe()->rightX, pe()->rightY, pe()->rightZ);
		  }
		glEnd();
		if (showHandle) {
			glBegin(GL_LINES);
			  glVertex3d(pe()->startX, pe()->startY, pe()->startZ);
			  glVertex3d(pe()->leftX, pe()->leftY, pe()->leftZ);
			  glVertex3d(pe()->rightX, pe()->rightY, pe()->rightZ);
			  glVertex3d(pe()->endX, pe()->endY, pe()->endZ);
			glEnd();
		}
		if (showHeight) {
			glBegin(GL_LINES);
			  glVertex3d(pe()->startX, pe()->startY, pe()->startZ);
			  glVertex3d(pe()->startX, 0, pe()->startZ);
			  glVertex3d(pe()->startX-.05, 0, pe()->startZ+.05);
			  glVertex3d(pe()->startX+.05, 0, pe()->startZ-.05);
			  glVertex3d(pe()->startX-.05, 0, pe()->startZ-.05);
			  glVertex3d(pe()->startX+.05, 0, pe()->startZ+.05);

			  glVertex3d(pe()->endX, pe()->endY, pe()->endZ);
			  glVertex3d(pe()->endX, 0, pe()->endZ);
			  glVertex3d(pe()->endX-.05, 0, pe()->endZ+.05);
			  glVertex3d(pe()->endX+.05, 0, pe()->endZ-.05);
			  glVertex3d(pe()->endX-.05, 0, pe()->endZ-.05);
			  glVertex3d(pe()->endX+.05, 0, pe()->endZ+.05);
			glEnd();
		}

		glColor3f(colorCenter[0], colorCenter[1], colorCenter[2]);
		glPointSize(4.0);
		glBegin(GL_POINTS);
		  glPointSize(4.0);
		  glVertex3d(pc()->startX, pc()->startY, pc()->startZ);
		  if (pc.isTail()) {
			  glVertex3d(pc()->endX, pc()->endY, pc()->endZ);
		  }
		  if (showHandle) {
			glVertex3d(pc()->leftX, pc()->leftY, pc()->leftZ);
			glVertex3d(pc()->rightX, pc()->rightY, pc()->rightZ);
		  }
		glEnd();
		if (showHandle) {
			glBegin(GL_LINES);
			  glVertex3d(pc()->startX, pc()->startY, pc()->startZ);
			  glVertex3d(pc()->leftX, pc()->leftY, pc()->leftZ);
			  glVertex3d(pc()->rightX, pc()->rightY, pc()->rightZ);
			  glVertex3d(pc()->endX, pc()->endY, pc()->endZ);
			glEnd();
		}
		if (showHeight) {
			glBegin(GL_LINES);
			  glVertex3d(pc()->startX, pc()->startY, pc()->startZ);
			  glVertex3d(pc()->startX, 0, pc()->startZ);
			  glVertex3d(pc()->startX-.05, 0, pc()->startZ+.05);
			  glVertex3d(pc()->startX+.05, 0, pc()->startZ-.05);
			  glVertex3d(pc()->startX-.05, 0, pc()->startZ-.05);
			  glVertex3d(pc()->startX+.05, 0, pc()->startZ+.05);

			  glVertex3d(pc()->endX, pc()->endY, pc()->endZ);
			  glVertex3d(pc()->endX, 0, pc()->endZ);
			  glVertex3d(pc()->endX-.05, 0, pc()->endZ+.05);
			  glVertex3d(pc()->endX+.05, 0, pc()->endZ-.05);
			  glVertex3d(pc()->endX-.05, 0, pc()->endZ-.05);
			  glVertex3d(pc()->endX+.05, 0, pc()->endZ+.05);
			glEnd();
		}

		// 1.5. put time label on points
		glListBase(Font);

		glColor3f(colorEye[0], colorEye[1], colorEye[2]);
		glRasterPos3f(pe()->startX, pe()->startY, pe()->startZ);
		sprintf(text, "%d", start);
		glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
		if (pe.isTail()) {
			glRasterPos3f(pe()->endX, pe()->endY, pe()->endZ);
			sprintf(text, "%d", end);
			glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
		}

		glColor3f(colorCenter[0], colorCenter[1], colorCenter[2]);
		glRasterPos3f(pc()->startX, pc()->startY, pc()->startZ);
		sprintf(text, "%d", start);
		glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
		if (pc.isTail()) {
			sprintf(text, "%d", end);
			glRasterPos3f(pc()->endX, pc()->endY, pc()->endZ);
			glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
		}
	}

	//-------------------------
	// 2. draw the current time
	self->SetTime(currentTime);
	glPointSize(6.0);
	glBegin(GL_POINTS);
	  glColor3f(colorEye[0], colorEye[1], colorEye[2]);
	  glVertex3d(currentEye()->X(currentTime), currentEye()->Y(currentTime), 
				 currentEye()->Z(currentTime));

	  glColor3f(colorCenter[0], colorCenter[1], colorCenter[2]);
	  glVertex3d(currentCenter()->X(currentTime), currentCenter()->Y(currentTime), 
				 currentCenter()->Z(currentTime));
	glEnd();
	if (showLine) {
		glBegin(GL_LINES);
		  glColor3f(colorEye[0], colorEye[1], colorEye[2]);
		  glVertex3d(currentEye()->X(currentTime), currentEye()->Y(currentTime), 
					 currentEye()->Z(currentTime));

		  glColor3f(colorCenter[0], colorCenter[1], colorCenter[2]);
		  glVertex3d(currentCenter()->X(currentTime), currentCenter()->Y(currentTime), 
					 currentCenter()->Z(currentTime));
		glEnd();
	}
		
	//-----------------------------------
	// 3. draw the position of the bodies

}

//------------------------------------------------------------------------------------------------

int ViewMovie::Pick(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const
{
GLuint selectBuf[SELECT_BUFSIZE], *ptr, names, min = 0xffffffff;
GLint hits;
GLint viewport[4];
int x, y, i, start, end, path = -1, time = -1, type = -1;
Tcl_Obj *res;


	// no picking if in play mode
	if (mode == PLAY || !currentEye) {
		return TCL_OK;
	}

	// try to pick an element
	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "x y");
		return TCL_ERROR;
	}

	// get the coordinate of the pick
	Tcl_GetIntFromObj(interp, objv[2], &x);
	Tcl_GetIntFromObj(interp, objv[3], &y);

	// enter in selection mode
    glSelectBuffer(SELECT_BUFSIZE, selectBuf);
    glRenderMode(GL_SELECT);

    glInitNames();
    glPushName(0);

	SetGLView(TRUE, x, y);


#define EYE 0
#define CENTER 1
#define POINT 0
#define LEFT_HANDLE 1
#define RIGHT_HANDLE 2
#define DRAW(time, what, X, Y, Z) 	glLoadName(time); glPushName(0); glLoadName(what); \
									glBegin(GL_POINTS); glVertex3d(X, Y, Z); glEnd(); \
									glPopName();


	// 1. draw only the points
	ListIterator<Bezier3D> pe, pc;
	for(pe.reset(eye), pc.reset(center); pe; ++pe, ++pc) {
		// 1.1. get parameters
		start = pe()->start;
		end = pe()->end;

		// path for EYE
		glLoadName(EYE);
		glPushName(0);

		glPointSize(4.0);
		DRAW(start, POINT, pe()->startX, pe()->startY, pe()->startZ);
		if (pe.isTail()) {
			DRAW(end, POINT, pe()->endX, pe()->endY, pe()->endZ);
		}
		if (showHandle) {
			DRAW(start, LEFT_HANDLE, pe()->leftX, pe()->leftY, pe()->leftZ);
			DRAW(end, RIGHT_HANDLE, pe()->rightX, pe()->rightY, pe()->rightZ);
		}
		glPopName();

		// path for CENTER
		glLoadName(CENTER);
		glPushName(0);

		glPointSize(4.0);
		DRAW(start, POINT, pc()->startX, pc()->startY, pc()->startZ);
		if (pe.isTail()) {
			DRAW(end, POINT, pc()->endX, pc()->endY, pc()->endZ);
		}
		if (showHandle) {
			DRAW(start, LEFT_HANDLE, pc()->leftX, pc()->leftY, pc()->leftZ);
			DRAW(end, RIGHT_HANDLE, pc()->rightX, pc()->rightY, pc()->rightZ);
		}
		glPopName();
	}
    glFlush();
    hits = glRenderMode(GL_RENDER);

	// 2. analyse the result
    ptr = selectBuf;
    for (i = 0; i < hits; i++) {
		if ((names = *ptr++) != 3) {
			Tcl_SetResult(interp, "[viewMovie::pick] internal error", NULL);
			return TCL_ERROR;
		}
		ptr++;	// skip the max
		if (*ptr < min) {
			min = *ptr++;
			path = *ptr++;
			time = *ptr++;
			type = *ptr++;
		} else {
			ptr += 4; // not good, skip it
		}
	}

	// 3. translate the result
	res = Tcl_NewListObj(0, NULL);	
	if (path != -1) {
		Tcl_ListObjAppendElement(interp, res,
				Tcl_NewStringObj((char *) (path == EYE ? "eye" : "center"), -1));
		Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(time));
		Tcl_ListObjAppendElement(interp, res, Tcl_NewStringObj(
				(char *) (type==POINT ? "point" : type==LEFT_HANDLE ? "lefthandle" : "righthandle"), -1));
	}

	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int ViewMovie::Info(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const
{
static char *cmdPtr[] = {
	"iscontrol", "preveye", "nexteye", "prevcenter", "nextcenter", (char *) 0};
enum {iscontrol, peye, neye, pcenter, ncenter} index;
int time = 0, min = 0, max = 1000000000;
Bezier3D *ptMin, *ptMax;
ListIterator<Bezier3D> p;
Tcl_Obj *res = 0;


	if (objc != 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "command time");
		return TCL_ERROR;
	}

	// get command
    if (Tcl_GetIndexFromObj(interp, objv[2], cmdPtr, "command", 0, (int *)&index) != TCL_OK) {
        return TCL_ERROR;
    }
	// get time
	Tcl_GetIntFromObj(interp, objv[3], &time);

	// do it
	if (index == iscontrol) {
		for(p.reset(eye); p; ++p) {
			if (time == p()->start || time == p()->end) {
				res = Tcl_NewBooleanObj(TRUE);
				break;
			}
		}
		if (!p) {
			res = Tcl_NewBooleanObj(FALSE);
		}
	} else {
		p.reset(index == peye || index == neye ? eye : center);
		ptMin = p();
		ptMax = 0;

		while (p) {
			if (p()->start < time && p()->end >= min) {
				min = p()->end;
				ptMin = p();
			}
			if (p()->start > time && p()->start <= max) {
				max = p()->start;
				ptMax = p();
			}
			++p;
		}

		res = Tcl_NewListObj(0, NULL);
		if (index == neye || index == ncenter) {
			if (!ptMax) {
				p.last(index == neye ? eye : center);
				ptMax = p();
				Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMax->endX));
				Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMax->endY));
				Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMax->endZ));
			} else {
				Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMax->startX));
				Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMax->startY));
				Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMax->startZ));
			}
		} else {
			Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMin->startX));
			Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMin->startY));
			Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(ptMin->startZ));
		}
	}


	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//================================================================================================

void ViewMovie::SetTime(int time)
{
	// if inside the current chunk
	if (currentEye()->start <= time && time < currentEye()->end) {
		return;
	} else if (time < currentEye()->start) {
		// find the right chunk, before the current one
		while (time < currentEye()->start) {
			if (currentEye.isHead()) {
				return;
			} else {
				--currentEye;
				--currentCenter;
				//--currentTwist;
			}
		}
	} else {
		// find the right chunk, after the current one
		while (time > currentEye()->end) {
			if (currentEye.isTail()) {
				currentTime = currentEye()->end; // force the end
				return;
			} else {
				++currentEye;
				++currentCenter;
				//++currentTwist;
			}
		}
	}
}


//================================================================================================

int ViewMovie_ObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
struct Togl *togl = (struct Togl *)clientData;
struct ViewMovie *view = (struct ViewMovie *) Togl_GetClientData(togl);
static char *optionsPtr[] = {
	"configure", "render", "swapbuffers", "makecurrent", 
	"get", "set", "persp", "ortho", "add", "remove", 
	"pick", "viewpoint", "record", "modify", "info", (char *) NULL};
enum {
	config, render, swap, current,
	get, set, persp, ortho, add, remove,
	pick, viewpt, record, modify, info} index;


	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "?options?");
		return TCL_ERROR;
	}

	if (Tcl_GetIndexFromObj(interp, objv[1], optionsPtr, "option", 0, (int *) &index) != TCL_OK) {
    	return TCL_ERROR;
	}

	Togl_MakeCurrent(togl);

	switch (index) {
		case config:
		case render:
		case swap:
		case current:	return Togl_DefaultProcessing(togl, interp, index, objc, objv);

		case add:		return view->Add(interp, objc, objv);
		case get:		return view->Get(interp, objc, objv);
		case info:		return view->Info(interp, objc, objv);
		case modify:	return view->Modify(interp, objc, objv);
		case pick:		return view->Pick(interp, objc, objv);
		case record:	return view->Record(interp, objc, objv);
		case remove:	return view->Remove(interp, objc, objv);
		case set:		return view->Set(interp, objc, objv);
		case viewpt:	return view->Viewpoint(interp, objc, objv);

		case ortho:		view->ortho = ORTHO;
						view->Redisplay();
						Togl_PostRedisplay(togl);
						break;
		case persp:		view->ortho = PERSP;
						view->Redisplay();
						Togl_PostRedisplay(togl);
						break;
	}

	return TCL_OK;
}
